/*
 *	emu.c
 *
 *	SG-1000/EMU
 *
 *	* Thu Jan 6 00:00:00 JST 2005 Naoyuki Sawa
 *	- 1st [XB
 */
#include "app.h"

/****************************************************************************
 *	
 ****************************************************************************/

/*{{fram.c*/
extern Z80 sg1000_cpu;
unsigned char z80_read(Z80* cpu, unsigned short addr);
void z80_write(Z80* cpu, unsigned short addr, unsigned char data);
/*}}fram.c*/

/****************************************************************************
 *	
 ****************************************************************************/

/*
 *	0x7f : SN76489
 *
 *	0xbe : TMS9918A Port#0
 *
 *	0xbf : TMS9918A Port#1
 *
 *	0xdc : 8255 Port#A
 *
 *		D0 : J1UP or KDB	(in, active-low)
 *		D1 : J1DN or KDB	(in, active-low)
 *		D2 : J1LF or KDB	(in, active-low)
 *		D3 : J1RI or KDB	(in, active-low)
 *		D4 : J1B1 or KDB	(in, active-low)
 *		D5 : J1B2 or KDB	(in, active-low)
 *		D6 : J2UP or KDB	(in, active-low)
 *		D7 : J2DN or KDB	(in, active-low)
 *	
 *	0xdd : 8255 Port#B
 *
 *		D0 : J2LF or KDB	(in, active-low)
 *		D1 : J2RI or KDB	(in, active-low)
 *		D2 : J2B1 or KDB	(in, active-low)
 *		D3 : J2B2 or KDB	(in, active-low)
 *		D4 : RESET		(in, active-low)
 *		D5 : PRN FAULT		(in, active-low)
 *		D6 : PRN BUSY		(in, active-low)
 *		D7 : CAS IN		(in, active-high)
 *	
 *	0xde : 8255 Port#C
 *
 *		D0 : JOY/KBD SEL	(out, 7=JOY/other=KEY)
 *		D1 : JOY/KBD SEL	(out, ..             )
 *		D2 : JOY/KBD SEL	(out, ..             )
 *		D3 : CAS OUT		(out)
 *		D4 : -			(out)
 *		D5 : PRN DATA		(out)
 *		D6 : PRN RESET		(out)
 *		D7 : PRN FEED		(out)
 *
 *	0xdf : 8255 Control
 *
 */

unsigned char
z80_in(Z80* cpu, unsigned short addr)
{
	addr &= 0xff;

	switch(addr) {
	//case 0x7f: /* SN76489 */

	case 0xbe: /* TMS9918A Port#0 */
	case 0xbf: /* TMS9918A Port#1 */
		return tms9918_read(&sg1000_vdp, addr & 1);

	case 0xdc: /* 8255 Port#A */
		if((sg1000_io[0xde] & 7) != 7) {
			return -1;
		}
		break;
	//case 0xdd: /* 8255 Port#B */
	//case 0xde: /* 8255 Port#C */
	//case 0xdf: /* 8255 Control */
	}

	return sg1000_io[addr];
}

void
z80_out(Z80* cpu, unsigned short addr, unsigned char data)
{
	addr &= 0xff;

	switch(addr) {
	case 0x7f: /* SN76489 */
		sn76489_write(&sg1000_dcsg, data);
		return;

	case 0xbe: /* TMS9918A Port#0 */
	case 0xbf: /* TMS9918A Port#1 */
		tms9918_write(&sg1000_vdp, addr & 1, data);
		return;

	//case 0xdc: /* 8255 Port#A */
	//case 0xdd: /* 8255 Port#B */
	//case 0xde: /* 8255 Port#C */
	//case 0xdf: /* 8255 Control */
	}

	sg1000_io[addr] = data;
}

void
sg1000_mix(short* wbuff)
{
	sn76489_mix(&sg1000_dcsg, wbuff, 320);
}

int
sg1000_stream_callback(short* wbuff, int param)
{
	/* FRAM4X^bNgpɂ֗āAUSB荞ݓdȂƁA
	 * FRAM4X^bNĂ܂ꍇ܂B
	 * ̂悤ȏ󋵂邽߂ɁASRAMX^bN֐؂ւ܂B
	 * sg1000_mix()̓X^bNˑȂ̂ŁASRAMX^bNł[łB
	 */
	static unsigned char stack[1024];
	stack_switch(stack + 1024, sg1000_mix, (int)wbuff, 0);
	return 0;
}

int
sg1000_run(const char* fname, int sound_on)
{
	int retval;
	int input;
	int cnt;
	FILE* fp;

	/********************************************************************
	 *	
	 ********************************************************************/

	memset(sg1000_mem, 0, 0x10000);
	memset(sg1000_io, -1, 0x100);

	fp = fopen(fname, "rb");
	if(!fp) {
		return -1;
	}
	fread(sg1000_mem, 1, 0x10000, fp);
	fclose(fp);

	z80_reset(&sg1000_cpu, z80_read, z80_write, z80_in, z80_out);
	tms9918_reset(&sg1000_vdp);
	sn76489_reset(&sg1000_dcsg, NTSC_COLOR_SUBCARRIER);
	if(sound_on) {
		stream_play(320, sg1000_stream_callback, 0, 0);
	}

	/********************************************************************
	 *	s
	 ********************************************************************/

	cnt = 0;
	for(;;) {
		fram4_call(z80_run, (int)&sg1000_cpu, NTSC_COLOR_SUBCARRIER / 60);
		cnt += FPS;
		if(cnt >= 60) {
			cnt -= 60;

			/* \XVB */
			retval = fram4_call(tms9918_update, (int)&sg1000_vdp, (int)vbuff);
			tms9918_reduce(vbuff, vbuff); /* X^bNˑȂ̂ŃX^bN̂܂ */
			schedule();

			/* ͍XV */
			if(joy & TRG_SELECT) {
				break;
			}
			input = -1;
			if(joy & PAD_UP) input &= ~(1 << 0);
			if(joy & PAD_DN) input &= ~(1 << 1);
			if(joy & PAD_LF) input &= ~(1 << 2);
			if(joy & PAD_RI) input &= ~(1 << 3);
			if(joy & PAD_B ) input &= ~(1 << 4);
			if(joy & PAD_A ) input &= ~(1 << 5);
			sg1000_io[0xdc] = input;

		} else {
			retval = tms9918_update(&sg1000_vdp, NULL); /* ςނ̂ŃX^bN̂܂ */
		}

		if(retval & 0x80/*Interrupt Flag*/) {
			z80_int(&sg1000_cpu, 0); /* ςނ̂ŃX^bN̂܂ */
		}
	}

	/********************************************************************
	 *	N[Abv
	 ********************************************************************/

	if(sound_on) {
		stream_stop();
	}

	return 0;
}

